/*
	Copyright (C) 2006 Ivo van Doorn

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the
	Free Software Foundation, Inc.,
	59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/laptopkey.h>

MODULE_AUTHOR("Ivo van Doorn <IvDoorn@gmail.com>");
MODULE_VERSION("1.0");
MODULE_DESCRIPTION("Laptop keys support");
MODULE_LICENSE("GPL");

/*
 * Default laptop_key_device structure.
 */
static struct laptop_key_device default_keydev = {
	.dev_name	= "laptopkey",
	.delay		= 100,
};

void laptopkey_key_event(struct laptop_key *key, int status)
{
	/*
	 * Check if the hardware needs to be informed of
	 * the status change.
	 */
	if (key->key_event)
		key->key_event(key->data, status);

	/*
	 * Send input event to userspace.
	 */
	input_report_key(key->keydev->input_dev, key->key, status);
	input_sync(key->keydev->input_dev);

}

static void laptopkey_poll(unsigned long data)
{
	struct laptop_key_device *keydev = (struct laptop_key_device *)data;
	struct laptop_key *key;
	struct list_head *entry;
	int status;

	spin_lock(&keydev->lock);
	list_for_each(entry, &keydev->list){
		key = list_entry(entry, struct laptop_key, entry);

		/*
		 * Check if this key requires polling.
		 * Otherwise just skip this key,
		 * and move to the next one.
		 */
		if (!key->key_poll)
			continue;

		/*
		 * Poll the status, and report a key event
		 * when the status has changed.
		 */
		status = !!key->key_poll(key->data);
		if (status != key->current_status) {
			laptopkey_key_event(key, status);
			key->current_status = status;
		}
	}
	spin_unlock(&keydev->lock);

	/*
	 * Check if the polling can continue.
	 */
	if (keydev->delay == 0)
		return;

	mod_timer(&keydev->timer, jiffies + msecs_to_jiffies(keydev->delay));
}

void laptopkey_dev_add_key(struct laptop_key_device *keydev,
			  struct laptop_key *key)
{
	/*
	 * If polling is enabled, obtain the initial
	 * status from the polling functions, otherwise
	 * set it to 0.
	 */
	if (key->key_poll)
		key->current_status = key->key_poll(key->data);
	else
		key->current_status = 0;

	/*
	 * Add key to key device.
	 */
	key->keydev = keydev;

	spin_lock(&keydev->lock);
	list_add(&key->entry, &keydev->list);
	spin_unlock(&keydev->lock);

	/*
	 * Add key to the input device.
	 */
	set_bit(key->key, keydev->input_dev->keybit);
}

void laptopkey_dev_del_key(struct laptop_key_device *keydev,
			   struct laptop_key *key)
{
	/*
	 * Remove list from device.
	 */
	spin_lock(&keydev->lock);
	list_del(&key->entry);
	spin_unlock(&keydev->lock);

	/*
	 * Remove key to the input device.
	 */
	clear_bit(key->key, keydev->input_dev->keybit);
}

void laptopkey_add_key(struct laptop_key *key)
{
	laptopkey_dev_add_key(&default_keydev, key);
}

void laptopkey_del_key(struct laptop_key *key)
{
	laptopkey_dev_del_key(&default_keydev, key);
}

int laptopkey_register_device(struct laptop_key_device *keydev)
{
	int status;

	/*
	 * Create the input device
	 */
	keydev->input_dev = input_allocate_device();
	if (!keydev->input_dev) {
		printk(KERN_ERR "Failed to allocate input device %s.\n",
			keydev->dev_name);
		return -ENOMEM;
	}

	/*
	 * Initialize the input device
	 */
	keydev->input_dev->name = "Laptop keys";
	keydev->input_dev->phys = keydev->dev_name;
	keydev->input_dev->id.bustype = BUS_HOST;

	/*
	 * Register the input device
	 */
	status = input_register_device(keydev->input_dev);
	if (status) {
		printk(KERN_ERR "Failed to register input device %s.\n",
			keydev->dev_name);
		input_free_device(keydev->input_dev);
		return status;
	}

	/*
	 * Initialize list head.
	 */
	INIT_LIST_HEAD(&keydev->list);

	/*
	 * Initialize lock.
	 */
	spin_lock_init(&keydev->lock);

	/*
	 * Initialize timer.
	 */
	init_timer(&keydev->timer);
	keydev->timer.function = laptopkey_poll;
	keydev->timer.data = (unsigned long)keydev;
	keydev->timer.expires = jiffies + msecs_to_jiffies(keydev->delay);
	add_timer(&keydev->timer);

	printk(KERN_INFO "Created new %s: %s.\n",
		keydev->input_dev->name, keydev->input_dev->phys);

	return 0;
}

void laptopkey_unregister_device(struct laptop_key_device *keydev)
{
	/*
	 * Clear the delay field before deleting the timer,
	 * to prevent the scheduled work to rearm itself.
	 */
	keydev->delay = 0;
	del_timer_sync(&keydev->timer);

	/*
	 * Remove input device.
	 */
	input_unregister_device(keydev->input_dev);
	input_free_device(keydev->input_dev);
}

static int __init radiobtn_init(void)
{
	printk(KERN_INFO "Loading radio button driver.\n");
	return laptopkey_register_device(&default_keydev);
}

static void __exit radiobtn_exit(void)
{
	laptopkey_unregister_device(&default_keydev);
	printk(KERN_INFO "Unloading radio button driver.\n");
}

EXPORT_SYMBOL(laptopkey_add_key);
EXPORT_SYMBOL(laptopkey_del_key);
EXPORT_SYMBOL(laptopkey_register_device);
EXPORT_SYMBOL(laptopkey_unregister_device);
EXPORT_SYMBOL(laptopkey_dev_add_key);
EXPORT_SYMBOL(laptopkey_dev_del_key);
EXPORT_SYMBOL(laptopkey_key_event);

module_init(radiobtn_init);
module_exit(radiobtn_exit);
